# Author: Robert Park <robru@gottengeography.ca>, (C) 2011
# Copyright: See COPYING file included with this distribution.
#
# Package version data.
#
# Please note that this file is equally valid as Python as it is Bash,
# as it is also imported by the po/initialize-pot.sh script.

APPNAME='GottenGeography'
PACKAGE='gottengeography'
VERSION='2.5'

AUTHOR='Robert Park'
EMAIL='robru@gottengeography.ca'
